@@ -1,4 +1,4 @@
-use std::time::Instant;
+use crate::Instant;
 
 #[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
 pub(crate) enum Timer {
@@ -18,10 +18,12 @@ pub(crate) enum Timer {
     Pacing = 6,
     /// When to invalidate old CID and proactively push new one via NEW_CONNECTION_ID frame
     PushNewCid = 7,
+    /// When to send an immediate ACK if there are unacked ack-eliciting packets of the peer
+    MaxAckDelay = 8,
 }
 
 impl Timer {
-    pub(crate) const VALUES: [Self; 8] = [
+    pub(crate) const VALUES: [Self; 9] = [
         Self::LossDetection,
         Self::Idle,
         Self::Close,
@@ -30,13 +32,14 @@ impl Timer {
         Self::KeepAlive,
         Self::Pacing,
         Self::PushNewCid,
+        Self::MaxAckDelay,
     ];
 }
 
 /// A table of data associated with each distinct kind of `Timer`
 #[derive(Debug, Copy, Clone, Default)]
 pub(crate) struct TimerTable {
-    data: [Option<Instant>; 8],
+    data: [Option<Instant>; 10],
 }
 
 impl TimerTable {
