@@ -18,14 +18,14 @@ use crate::{
 };
 
 /// Cryptography interface based on *ring*
-#[cfg(feature = "ring")]
-pub(crate) mod ring;
+#[cfg(any(feature = "aws-lc-rs", feature = "ring"))]
+pub(crate) mod ring_like;
 /// TLS interface based on rustls
-#[cfg(feature = "rustls")]
+#[cfg(any(feature = "rustls-aws-lc-rs", feature = "rustls-ring"))]
 pub mod rustls;
 
 /// A cryptographic session (commonly TLS)
-pub trait Session: Send + 'static {
+pub trait Session: Send + Sync + 'static {
     /// Create the initial set of keys given the client's initial destination ConnectionId
     fn initial_keys(&self, dst_cid: &ConnectionId, side: Side) -> Keys;
 
@@ -127,7 +127,6 @@ pub trait ServerConfig: Send + Sync {
         &self,
         version: u32,
         dst_cid: &ConnectionId,
-        side: Side,
     ) -> Result<Keys, UnsupportedVersion>;
 
     /// Generate the integrity tag for a retry packet
@@ -146,7 +145,7 @@ pub trait ServerConfig: Send + Sync {
 }
 
 /// Keys used to protect packet payloads
-pub trait PacketKey: Send {
+pub trait PacketKey: Send + Sync {
     /// Encrypt the packet payload with the given packet number
     fn encrypt(&self, packet: u64, buf: &mut [u8], header_len: usize);
     /// Decrypt the packet payload with the given packet number
@@ -166,7 +165,7 @@ pub trait PacketKey: Send {
 }
 
 /// Keys used to protect packet headers
-pub trait HeaderKey: Send {
+pub trait HeaderKey: Send + Sync {
     /// Decrypt the given packet's header
     fn decrypt(&self, pn_offset: usize, packet: &mut [u8]);
     /// Encrypt the given packet's header
